/*
 * StatDomTree.java
 * - Modified by Hiroki Arimura on 18 MAR 03
 *   Implementation with JAXP
 *   To deal with HTML files, use tidy --output-xml option.
 * - Created by Hiroki Arimura on 10 DEC 01
 * - based on Mattree10.java by Kenji Abe, 2001 for our SDM paper
*/

/*
   This software may be freely used for academic purpose only. Please
   contact us for profit use. However, when distributed, the original
   source must be clearly stated, and, when the source code is
   distributed, the copiright notice must be retained and any
   alterations in the code must be clearly marked. No warranty is
   given regarding the quality of this software.
*/

import java.io.*;
import java.lang.*;
import java.util.*;
import java.math.*;
import java.text.*;

//For new implementation with jaxp (sun java2)
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.w3c.dom.*;

//for jaxp1.2 attached to java2_sdk1.4.1
import org.apache.crimson.tree.*;  //For UserObject() of class ElementEx

public class DomWrapper {

    /************************************************************
     * Global Compilation Switch
     ************************************************************/
    static boolean ENABLE_ATTACH_ATTR  = false;
    static boolean ENABLE_PRINT_TREE   = true;
    static final boolean DOM_VARIDATION      = false; //to be false

    static final boolean DEBUG_TAGTABLE  = false;
    static final boolean DEBUG_ATTRTABLE = false;
    static final boolean REPORT_TREESTAT  = false;
    static final boolean DEBUG_TAGNAMES = false;
    static final boolean DEBUG_ATTRNAMES= false;

    static void setEnableAttachAttr(boolean state) {
	ENABLE_ATTACH_ATTR = state;	
    }

    /************************************************************
     * Global variables: they should be non-static
     ************************************************************/
    static Document gDocument = null;    //Dom document for an input file
    static Node     gDomRoot  = null; 

    /************************************************************
     * Statistics
     * 
     ************************************************************/
    //Tag Statistics
    static SymbolTable   tagTable  = null; //A tag table

    //Attribute Statistics
    static StringPairMap attrTable = null; //A attribute table
  
    //Tree Statistics
    static boolean isStatAvailable = false;
    static long num_nodes = 0;	
    static long max_depth = 0;	
    static long sum_depth = 0;	
    static long num_leaves = 0;
    static long sum_branch = 0;	
    static long max_branch = 0;

    /**********************************************************************
     * Dom Wrapper for OpenXML
     * createDom: 
     * - Parse a given input HTML File and then returns the root node
     * - Use instead of YmUtil.parseByOpenXML in YmUtil.java
     **********************************************************************/

    public static Node createDom(String inFileName) throws Exception {
	try {
	    //The factory pattern:
	    //  to be independent from a particular XML parser.
	    DocumentBuilderFactory dbfactory
		= DocumentBuilderFactory.newInstance();
	    dbfactory.setValidating(false);
	    DocumentBuilder builder = dbfactory.newDocumentBuilder();
	    Document d = gDocument = builder.parse(inFileName);

	    //Main
	    gDomRoot = d.getDocumentElement();

	    RemoveEmptyTextNodes.removeWhiteSpaceNodes(gDomRoot);

	    if (ENABLE_ATTACH_ATTR) {
		transformTree(gDomRoot);
	    }

	} catch(Exception e) {
	    e.printStackTrace();
	}
	return gDomRoot;
    }

    public static Node createDom(File inputFile) throws Exception {
	return createDom(inputFile.toString());
    }

    public static Document getDocument() { return gDocument; }

    public static Node getRoot() { return gDomRoot; }

    public static long getTreeSize() {
	if (!isStatAvailable)
	    DomWrapper.statTree(gDomRoot);	  
	return num_nodes;
    }

    /**********************************************************************
     * Create extended Dom tree
     * - By attaching attribute-value pairs as element nodes
     * - Attribute nodes appear from left to right in the lexicographic
     * order of their names.
     **********************************************************************/

    /** Main routine for the conversion */
    public static void transformTree(Node node) 
	{
	    String name = node.getNodeName();
	    appendAttributeNodes(node);

	    for (Node next = node.getFirstChild(); next != null;
		 next = next.getNextSibling()){
		transformTree(next);
	    }
	}

    /** Utility: 
     *  - insert a new child at the first position */
    static void insertFirstNode(Node node, Node newChild) {
	node.insertBefore(newChild, node.getFirstChild());
    }

    /** Sub-routine for the conversion: 
     * - Sort the attributes of a given node */
    static void appendAttributeNodes(Node node)
	{
	    String name = node.getNodeName();

	    if (DEBUG_ATTRNAMES) { //For debugging
		Debug.println("\t@elem: " + name); 
	    }//DEBUG_TAGNAMES

	    if (node.getNodeType() == Node.ELEMENT_NODE) {

		//Get a set of attributes for elem
		NamedNodeMap map =
		    ((Element)node).getAttributes(); //Downward type casting

		ArrayList attrnameList = new ArrayList();
		for (int i = 0; i < map.getLength(); i++) {
		    Attr attr = (Attr) map.item(i);
		    attrnameList.add(attr.getNodeName());
		}

		//Sort attributes by their names
		Object sortedNames[] = attrnameList.toArray(); 
		Arrays.sort(sortedNames);
		for (int i = 0; i < sortedNames.length; i++) {
		    String attrName = (String) sortedNames[i]; 
		    String attrValue = map.getNamedItem(attrName).getNodeValue();

		    Element attrNode = gDocument.createElement(attrName);
		    attrNode.setNodeValue(attrValue);
		    UserData.setVirtualAttrValue(attrNode, attrValue);

		    node.appendChild(attrNode);

		    if (DEBUG_ATTRNAMES) 
			Debug.println("\t\tattr: " + attrName +
				      "\tvalue: " + attrValue);
		}
	    }
	}

    
    public static void initialize()
	{
	    tagTable  = new SymbolTable();   //A tagname table
	    attrTable = new StringPairMap(); //An attribute table 

	    num_nodes = 0;
	    num_leaves= 0;
	    max_depth = 0;
	    sum_depth = 0;
	    max_branch = 0;
	    sum_branch = 0;
	}

    public static void report()
	{
	    if (DEBUG_TAGTABLE == true) {
		Debug.println("Reporting tagTable...");
		tagTable.reportCount();
	    }
	    if (DEBUG_ATTRTABLE == true) {
		Debug.println("Reporting attrTable...");
		attrTable.reportCount();
	    }

	    if (REPORT_TREESTAT == true) {
		Debug.println("@stat Num of nodes  = " + num_nodes);
		Debug.println("@stat Num of leaves = " + num_leaves);
		Debug.println("@stat Num of tags   = " + tagTable.getSize());
		Debug.println("@stat Depth_average  = " +  sum_depth/num_leaves +
			      "\t Depth_maximum = " +  max_depth);
		long num_internals = (num_nodes - num_leaves); 
		Debug.println("@stat Branch_average = " +  sum_branch/num_internals +
			      "\t Branch_maximum = " +  max_branch);
	    }
	}

    public static String getStatString() {
	long num_internals = (num_nodes - num_leaves); 
	String str = "@stat Num of nodes  = " + num_nodes+
	    "\n@stat Num of leaves = " + num_leaves+
	    "\n@stat Num of tags   = " + tagTable.getSize()+
	    "\n@stat Depth_average  = " +  sum_depth/num_leaves +
	    "\t Depth_maximum = " +  max_depth+
	    "\n@stat Branch_average = " +  sum_branch/num_internals +
	    "\t Branch_maximum = " +  max_branch+"\n";
	return str;
    }

    public static void statTree(Node node) 
	{
	    System.out.println("DomWrapper.statTree: computing stat...");
	    tagTable.clear();
	    attrTable.clear();
	    statTree0(node, 0);
	    isStatAvailable = true;
	}

    static void registerAttributes(Node node, StringPairMap attrTable)
	{
	    String name = node.getNodeName();
	    if (DEBUG_TAGNAMES) { //For debugging
		Debug.println("@node name: " + name +
			      "\tnode type: "+ Node.ELEMENT_NODE);
	    }//DEBUG_TAGNAMES

	    if (node.getNodeType() == Node.ELEMENT_NODE) {
		Element elem = (Element) node;

		//Get a set of attributes for elem
		NamedNodeMap map = elem.getAttributes();

		//For each attr, do the followings
		for (int i = 0; i < map.getLength(); i++) {
		    Attr attr = (Attr) map.item(i);

		    //Get attribute name 
		    String attrname = attr.getNodeName();

		    //Get attribute value
		    String attrvalue = attr.getNodeValue();

		    //Do something with (attrname, attrvalue)
		    int c = attrTable.addPair(attrname, attrvalue);

		    //For debugging
		    if (DEBUG_TAGNAMES) {
			int c0 = attrTable.getSounceCount(attrname);
			Debug.println("\t" + c0 + ":" + c + ":" + 
				      "\tattrname=["+ attrname +"]" +
				      "\tattrvalue=["+attrvalue+"]");
		    }//DEBUG_TAGNAMES

		}
	    }
	} 

    private static void statTree0(Node node, int depth) 
	{
	    String name = node.getNodeName(); 
	    int nid = tagTable.register(name);
	    registerAttributes(node, attrTable);
	    //appendAttributeNodes(node);

	    num_nodes++;

	    if  (node.hasChildNodes()){ 
		int branch = 0; 

		for (Node nextnode = node.getFirstChild();
		     nextnode != null;
		     nextnode = nextnode.getNextSibling())
		    {
			statTree0(nextnode, depth + 1);
			branch = branch + 1;
		    }
		max_branch = Math.max(max_branch, branch);
		sum_branch = sum_branch + branch;
	    }
	    else { 
		num_leaves = num_leaves + 1; 
		max_depth = Math.max(max_depth, depth);
		sum_depth = sum_depth + depth;
	    }
	}

    public static Node pthParent(Node node, int p)
	{ //Note: node is w3c.dom.Node type 
	    while (p > 0){
		node = node.getParentNode();
		p--;
	    }
	    return node;
	}

    /************************************************************
     * Main
     * - A test routine for traversing a DOM tree
     ************************************************************/

    public static void main(String args[]){

	// Process arguments
	//System.out.println("##Arg Length = " + args.length);
	int argc = 0;
	File pageFile = null;

	try {
	    if (args.length < 1) {
		System.out.println("DomWrapper: too few/many arguments. ");
		System.out.println("Usage: DomWrapper htmlfile");
		System.exit(1);
	    }

	    String filename = args[argc++];
	    pageFile = new File(filename);
	    System.out.println("##File name = " + filename);
	    System.out.println("##File size (byte) = " + pageFile.length());

	    while (argc < args.length){
		if (args[argc].equals("-attr")){
		    argc++;
		    ENABLE_ATTACH_ATTR = true;
		}
		else
		    throw new IllegalArgumentException();
	    }
	}
	catch(IllegalArgumentException e){
	    System.err.println("Too few/many arguments or wrong options.");
	    System.out.println("Usage: DomWrapper htmlfile");
	    System.exit(1);
	}


	try{
	    long cputime;
	    Debug.task("Initialization...");
	    initialize();

	    Debug.task("Building an DOM tree by reading the input HTML/XML file...");
	    cputime = Debug.checkin();
	    Node root    = gDomRoot = DomWrapper.createDom(pageFile);
	    Document doc = DomWrapper.getDocument();
	    Debug.checkout(cputime);

	    Debug.task("Report statistics of the original DOM tree...");
	    DomWrapper.statTree(gDomRoot);
	    DomWrapper.report();
	    Debug.task("END orig");

	    if (ENABLE_ATTACH_ATTR) {
		Debug.task("Building extended DOM tree...");
		//DomWrapper.createAttrTree(root, doc);
		DomWrapper.transformTree(root);
		//Bug: The crimson perser cannot allow some invalid chars in elements
		RemoveEmptyTextNodes.removeWhiteSpaceNodes(gDomRoot);

		Debug.task("Report statistics of the preprocessed DOM tree...");
		DomWrapper.statTree(gDomRoot);
		DomWrapper.report();
	    }
	    if (ENABLE_PRINT_TREE == true) {
		Debug.task("Printing the dom tree...");
		//printTree(root);
		PrintDomTree.printTree(gDomRoot, doc);
	    }

	    tagTable = null; 
	    Debug.task("Done.");

	}

	catch(Exception e){
	    System.out.println(e);
	}
    }

}

class StringPairMap extends HashMap {
  /** A two level string hash dictionary for pairs of strings (source, dest)
   */
  HashMap sourceMap;
  static String ATTRCOUNT = "@self";

  public StringPairMap()
  {
    super();
    sourceMap = new HashMap();
  }

  private int addDest(HashMap destMap, String dest)
  {
    //Getting destMap
    if (destMap.containsKey(dest) == false) 
      destMap.put(dest, new Integer(0));
    Integer c = (Integer) destMap.get(dest);
    int cnew  = c.intValue() + 1; 
    destMap.put(dest, new Integer(cnew)); //Increment
    return cnew; 
  }

  public int addPair(String source, String dest)
  {
    //Getting sourceMap
    if (this.sourceMap.containsKey(source) == false) 
      this.sourceMap.put(source, new HashMap());
    HashMap destMap = (HashMap) this.sourceMap.get(source);

    //Getting destMap
    int snew = addDest(destMap, StringPairMap.ATTRCOUNT);
    int cnew = addDest(destMap, dest);
    //if (destMap.containsKey(dest) == false) 
    //  destMap.put(dest, new Integer(0));
    //Integer c = (Integer) destMap.get(dest);
    //int cnew  = c.intValue() + 1; 
    //destMap.put(dest, new Integer(cnew)); //Increment
    return cnew; 
  }

  public int getSounceCount(String source)
  { /** Returns whether the edge (source, dest) is contained in
     */
    if (this.sourceMap.containsKey(source)){
      HashMap destMap = (HashMap) this.sourceMap.get(source);
      if (destMap.containsKey(StringPairMap.ATTRCOUNT))
	return ((Integer) destMap.get(StringPairMap.ATTRCOUNT)).intValue();
      else
	return 0;
    }
    else {
      return 0; 
    }
  }

  public int getPairCount(String source, String dest)
  { /** Returns whether the edge (source, dest) is contained in
     */
    if (this.sourceMap.containsKey(source)){
      HashMap destMap = (HashMap) this.sourceMap.get(source);
      if (destMap.containsKey(dest))
	return ((Integer) destMap.get(dest)).intValue();
      else
	return 0;
    }
    else {
      return 0; 
    }
  }

  public HashMap getSourceMap()
  {
    return (HashMap) this.sourceMap; 
  }

  public HashMap getDestMap(String source)
  {
    if (this.sourceMap.containsKey(source)){
      return (HashMap) this.sourceMap.get(source);
    }
    else
      return null;
  }

  public void reportCount()
  {
    HashMap sourceMap = (HashMap) this.sourceMap; 

    //Process sourceMap
    Set entries = sourceMap.entrySet();
    Iterator iter = entries.iterator();
    while (iter.hasNext()) {
      Map.Entry entry = (Map.Entry) iter.next();
      String source    = (String) entry.getKey();
      //Do something
      Debug.println("\tsource: " + source); 

      HashMap destMap = (HashMap) entry.getValue();

      //Process destMap
      Set d_entries = destMap.entrySet();
      Iterator d_iter = d_entries.iterator();
      while (d_iter.hasNext()) {
	Map.Entry d_entry = (Map.Entry) d_iter.next();
	String dest  = (String) d_entry.getKey();
	int count = ((Integer) d_entry.getValue()).intValue();
	//Do something
	Debug.println("\t" + count + "\tdest: " + dest);
      } // while dest
    } //while source
  }

}//StringPairMap


class PrintDomTree {

    private static Element gAttrNode = null;
    private static Document gDocument = null;

    static void printTree(Node node, Document doc) {
	gDocument = doc;
	Stack path = new Stack(); //Stack of nodes
	printTree0(node, path);
    }

    private static void printTree0(Node node, Stack path) {
	path.push(node);

	//Printing the element
	printPath(path);
	printNode(node);

	//Printing its attribute children
	printAtributeNodes(node, path);

	//Printing its element children
	Node child= node.getFirstChild();
	while (child!= null) {
	    printTree0(child, path);
	    child= child.getNextSibling();
	}
	path.pop();
    }

    static Element getDammyAttrNode() {
	if (gAttrNode == null) {
	    if (gDocument == null) {
		System.out.println("no document allocated!");
		System.exit(1);
	    }
	    gAttrNode = gDocument.createElement("dammyAttribute");
	}
	return gAttrNode;
    }

    public static void printAtributeNodes(Node node, Stack path) {
	if (node.getNodeType() != Node.ELEMENT_NODE)
	    return;

	//Get a set of attributes
	Element elem = (Element) node; //Downward type casting
	NamedNodeMap map = elem.getAttributes();
	ArrayList attrnameList = new ArrayList();

	//Store attribute nodes into a list;
	for (int i = 0; i < map.getLength(); i++) {
	    Attr attr = (Attr) map.item(i);
	    String attrname = attr.getNodeName();	//Get attribute name 
	    String attrvalue = map.getNamedItem(attrname).getNodeValue();

	    //Printing attribute name and value
	    path.push(getDammyAttrNode());
	    printPath(path);
	    System.out.println("@"+attrname+"=\""+attrvalue.trim()+"\"");
	    path.pop();
	}
    }
    private static void printPath(Stack path) {
	int len = path.size();
	for (int i = 0; i < len; i++) {
	    Node parent = (Node) path.get(i);
	    if (i < len-1) {
		if (parent.getNextSibling() == null) { //rightmost
 		    System.out.print("   "); //rightmost
		} else {
 		    System.out.print("|  ");
		}
	    }
	    else {
		if (parent.getNextSibling() == null) { //rightmost
		    System.out.print("`--"); //rightmost
		} else {
 		    System.out.print("|--");
		}
	    }
	}
    }

    private static void printNode(Node node) {
	switch (node.getNodeType()) {
	case Node.ELEMENT_NODE: {
	    System.out.print(""+""+node.getNodeName());
	    if (UserData.isVirtualAttrNode(node)) 
		UserData.printVirtualAttrNode(node);
	    else 
		System.out.println();
	    break;
	}
	case Node.ATTRIBUTE_NODE: {
	    System.out.println("AN: "+"\""+normalizeText(node.getNodeValue()).trim()+"\"");
	    break;
	}
	case Node.TEXT_NODE: {
	    System.out.println("#TEXT: "+"\""+normalizeText(node.getNodeValue()).trim()+"\"");
	    break;
	}
	case Node.PROCESSING_INSTRUCTION_NODE: {
	    System.out.println("PI: "+"\""+normalizeText(node.getNodeValue())+"\"");
	    break;
	}
	case Node.COMMENT_NODE: {
	    System.out.println("COMMENT: "+"\""+normalizeText(node.getNodeValue())+"\"");
	    break;
	}
	default: {
	    System.out.println("OTHERTYPE: "+node.getNodeType());
	    break;
	}
        }//switch
    }

    public static String normalizeText(String text) {
	if (text == null)
	    return null;
	StringBuffer buf = new StringBuffer();
	for (int i = 0; i < text.length(); i++) {
	    char C = text.charAt(i);
	    switch (C) {
	    case '\t': case ' ':
		{
		    buf.append(' ');
		    break;
		}
	    case '\n': case '\r': case '\f': case '\b':
		{
		    buf.append("\\n");
		    break;
		}
	    default:
		{
		    buf.append(C);
		    break;
		}
	    }
	}
	return buf.toString();
    }
}

class RemoveEmptyTextNodes {
    static void removeWhiteSpaceNodes(Node node) {
	Node nextnode= node.getFirstChild();
	while (nextnode!= null) {
	    Node child = nextnode;
	    nextnode= child.getNextSibling();

	    if (child.getNodeType() == Node.TEXT_NODE) {
		//For a text node
		if (isWhitespaceString(child.getNodeValue())) {
		    //Remove ignorable whitespace
		    node.removeChild(child);
		}
	    }
	    else if (child.getNodeType() == Node.ELEMENT_NODE) {
		//For an element node
		removeWhiteSpaceNodes(child);
	    }
	}
    }

    private static boolean isWhitespaceString(String text) {
	for (int i = 0; i < text.length(); i++) {
	    if (!Character.isWhitespace(text.charAt(i)))
		return false;
	}
	return true;
    }

}


class UserData {
    static final short UNDEF = -1;

    //Field
    short  nodeType  = UNDEF;
    String nodeValue = null;

    UserData() {
	nodeType  = UNDEF;
	nodeValue = null;
    }

    UserData(short type, String value) {
	nodeType  = type;
	nodeValue = value;
    }

    //get
    //long getNodeId() { return nodeId; }
    short getNodeType() { return nodeType; }
    String getNodeValue() { return nodeValue; }

    //set
    //void setNodeId(short id) { nodeId = id; }
    void setNodeType(short type) { nodeType = type; }
    void setNodeValue(String value) { nodeValue = value; }

    //Service routines:
    //

    static boolean isVirtualAttrNode(Node node) {
	if (node.getNodeType() != Node.ELEMENT_NODE)
	    return false;
	UserData udata = null;
	ElementEx elemex = (ElementEx)node; //Note: Crimson DOM specific code
	if ((udata = (UserData)elemex.getUserObject()) == null) 
	    return false;
	if (udata.getNodeType() == Node.ATTRIBUTE_NODE)
	    return true;
	else
	    return false;
    }

    static boolean hasVirtualValue(Node node) {
	if (node == null)
	    Util.errorExit("hasVirtualValue: null node!");
	if (node.getNodeType() == Node.TEXT_NODE)
	    return true;
	else if (node.getNodeType() == Node.ELEMENT_NODE) {
	    UserData udata = null;
	    ElementEx elemex = (ElementEx)node; //Note: Crimson DOM specific
	    if ((udata = (UserData)elemex.getUserObject()) == null) 
		return false;
	    else if (udata.getNodeType() == Node.ATTRIBUTE_NODE)
		return true;
	    else
		return false;
	}
	else {
	    Util.errorExit("hasVirtualValue: unexpected node type!");
	    return false;
	}
    }

    static String getVirtualValue(Node node) {
	String value;
	if (node == null)
	    Util.errorExit("getVirtualValue: null node!");
	if (node.getNodeType() == Node.TEXT_NODE ||
	    node.getNodeType() == Node.ELEMENT_NODE) {
	    return node.getNodeValue(); 
	}
	return null;
    }

    //Service routines
    static String getVirtualAttrValue(Node node) {
	ElementEx elemex = (ElementEx)node; //Note: Crimson DOM specific code
	UserData udata;
	if ((udata = (UserData)elemex.getUserObject()) == null) {
	    System.out.println("getVirtualAttrValue: null UserData!");
	    return null;
	}
	if (udata.getNodeType() != Node.ATTRIBUTE_NODE)
	    return null;
	else
	    return udata.getNodeValue(); 
    }

    static void setVirtualAttrValue(Node node, String value) {
	ElementEx elemex = (ElementEx)node; //Note: Crimson DOM specific code
	UserData udata;
	if ((udata = (UserData)elemex.getUserObject()) == null) {
	    udata = new UserData();
	    elemex.setUserObject(udata);
	}
	udata.setNodeType(Node.ATTRIBUTE_NODE); 
	udata.setNodeValue(value); 
    }

    static void printVirtualAttrNode(Node node) {
	String attrValue; 
	if ((attrValue = UserData.getVirtualAttrValue(node)) == null)
	    attrValue = "*"; 
	System.out.println("=\""+attrValue+"\"");
    }

}

